/*
"The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the content of this file.

The Initial Developer of the Original Code is Dr. Robert Kofler.
Carrer del Baluard; 73 21a
08003 Barcelona
All Rights Reserved.

Contributor(s): Dr. Robert Kofler.

Alternatively, the contents of this file may be used under the terms
of the _____ license (the  "[___] License"), in which case the
provisions of [______] License are applicable instead of those
above. If you wish to allow use of your version of this file only
under the terms of the [____] License and not to allow others to use
your version of this file under the MPL, indicate your decision by
deleting the provisions above and replace them with the notice and
other provisions required by the [___] License. If you do not delete
the provisions above, a recipient may use your version of this file
under either the MPL or the [___] License."
*/
using System.Collections;
using System.IO;
using System.Collections.Generic;
using System.Text;
using Bio.Seq;
using Bio.Alignment.Pairwise;
using System;

namespace Bio
{

    namespace Repetitive
    {

        public class LogRepetitiveUtility
        {

            private class NormalisingWindow
            {
                private int countValid = 0;
                private int summaryScore = 0;
                private int windowSize;
                public NormalisingWindow(int normalisingWindowSize)
                {
                    this.windowSize = normalisingWindowSize;
                }

                public void AddLogScore(byte toAdd)
                {
                    if (toAdd != 254)
                    {
                        countValid++;
                        summaryScore += toAdd;
                    }
                }

                public void RemoveLogScore(byte toRemove)
                {
                    if (toRemove != 254)
                    {

                        summaryScore -= toRemove;
                        countValid--;
                    }
                }

                public byte GetScore()
                {
                    if ((((float)countValid) / ((float)windowSize)) < 0.3333F) return (byte)254;
#if DEBUG


                        double round = ((double)summaryScore) / ((double)countValid);

#endif
                    return (byte)Math.Round(((double)summaryScore) / ((double)countValid), 0);
                }
            }

            /// <summary>
            /// Normalise the logodd scores along the sequence using a sliding window approach
            /// </summary>
            /// <param name="sequence"></param>
            /// <param name="normaliserWindowSize"></param>
            /// <returns></returns>
            public static ISequenceContainer NormaliseSequence(ISequenceContainer sequence, int normaliserWindowSize)
            {

                ///Initialize the normaliser 
                NormalisingWindow nm = new NormalisingWindow(normaliserWindowSize);
                int positionOffset_5p = normaliserWindowSize / 2;
                int positionOffset_3p = normaliserWindowSize - positionOffset_5p;
                ISequenceContainer normalisedLogOdd = SequenceFactory.GetDefaultSequence();

                //Initialize normaliser
                for (int i = 0; i < normaliserWindowSize; i++)
                {
                    nm.AddLogScore((byte)sequence[i]);
                }
                //initialize the sequence
                for (int i = 0; i < positionOffset_3p; i++)
                {
                    normalisedLogOdd.Append(nm.GetScore());
                }


                //iterate the sequence
                for (int i = normaliserWindowSize; i < sequence.Length; i++)
                {
                    nm.AddLogScore((byte)sequence[i]);
                    nm.RemoveLogScore((byte)sequence[i - normaliserWindowSize]);
                    normalisedLogOdd.Append(nm.GetScore());

                }
                //Fill up to the end
                for (int i = 0; i < positionOffset_5p; i++)
                {
                    normalisedLogOdd.Append((byte)nm.GetScore());
                }


                if (normalisedLogOdd.Length != sequence.Length) throw new DataMisalignedException("the logodd sequence does not have the same size than the original sequence");
                return normalisedLogOdd;
            }


        }



        /// <summary>
        /// Represents an index indicating the repetitivenes of a certain DNA sequence
        /// </summary>
        public class RepetitiveSequenceIndexInt
        {
            private int wordSize;
            private DNADictionary<int> index;


            /// <summary>
            /// Create an hashindex of the sequence which counts all entries in the hashtable to get an impression about the
            /// repetitive fraction in the sequence 
            /// </summary>
            /// <param name="wordSize"></param>
            /// <param name="indexValidator">is the hash-key valid, does not check if the entire sequence is valid, only if the small sequence of size wordsize, should be used as hash key</param>
            public RepetitiveSequenceIndexInt(int wordSize)
            {
                this.wordSize = wordSize;
                this.index = new DNADictionary<int>(wordSize);

            }

            public RepetitiveSequenceIndexInt(DNADictionary<int> repetitiveIndex, int wordsize)
            {
                this.index = repetitiveIndex;
                this.wordSize = wordsize;
    
            }

  

            /// <summary>
            /// Add a new Sequence to the index
            /// </summary>
            /// <param name="sequence"></param>
            public void AddToIndex(ISequenceContainer sequence)
            {
                   for (int i = 0; i < sequence.Length-wordSize; i++)
                {
                    ISequenceContainer seq = sequence.SubSequence(i,wordSize);

                    if (index.ContainsKey(seq))
                    {
                        this.index[seq]++;
                    }
                    else index.Add(seq, 1);
  
                }

            }


            /// <summary>
            /// Get the DNADictionary which was created for this special instance
            /// </summary>
            public DNADictionary<int> Dictionary
            {
                get
                {
                    return index;
                }
            }

            /// <summary>
            /// A repetitive index does only allow for a certain wordsize.
            /// Mixing of the word sizes is not valid. Retrieve the wordsize used in this repetitive index
            /// </summary>
            public int Wordsize
            {
                get
                {
                    return this.wordSize;
                }
            }

            /// <summary>
            /// Get the counts for a certain word
            /// </summary>
            /// <param name="word"></param>
            /// <returns></returns>
            public int? this[ISequenceContainer word]
            {
                get
                {
                    if (index.ContainsKey(word)) return index[word];
                    else return null;
                }
            }







            #region Converters


            /// <summary>
            /// Create the sequenceRepresenting the logOddRepetitiveSequence
            /// </summary>
            /// <param name="sequence"></param>
            /// <param name="generalInfo"></param>
            /// <returns></returns>
            public ISequenceContainer ConvertToLogOdd(ISequenceContainer sequence, INucSeqInfo generalInfo)
            {

                double expectedCount = ((double)generalInfo.Count_ATUCG / Math.Pow(4, wordSize));
                ISequenceContainer logodd = SequenceFactory.GetDefaultSequence();

                ///logOdd=lo
                /// lo=0 -> not overrepresented
                /// lo=1 -> 10 fold overrepresented
                /// lo=2 -> 100 fold overrpresented
                /// etc
                for (int i = 0; i < sequence.Length - wordSize+1; i++)
                {
                    ISequenceContainer s = sequence.SubSequence(i, wordSize);
                    if (index.ContainsKey(s))
                    {
                        double or=(double)index[s]/expectedCount;
                        if(or<1) or=1.0;
                        byte ap  = (byte)(Math.Round(Math.Log(or, 10.0), 1)*10);
#if DEBUG
                        int ind = index[s];
                        double log = Math.Log((double)index[s] / expectedCount, 10.0);
                        double round = (Math.Round(log, 1)*10);
                        if (ap == (byte)255) throw new Exception("crap");

#endif
                        logodd.Append((byte)ap);
                    }
                    else
                    {
                        logodd.Append((byte)254);
                    }
                }
                for (int i = 0; i < wordSize-1; i++)
                {
                    logodd.Append((byte)254);
                }

                if (sequence.Length != logodd.Length) throw new DataMisalignedException("the logodd sequence does not have the same size than the original sequence");
                return logodd;

            }


            #endregion

        }

        /// <summary>
        /// Represents an index indicating the repetitivenes of a certain DNA sequence
        /// </summary>
        public class RepetitiveSequenceIndexUshort
        {
            private int wordSize;
            private DNADictionary<ushort> index;


            /// <summary>
            /// Create an hashindex of the sequence which counts all entries in the hashtable to get an impression about the
            /// repetitive fraction in the sequence 
            /// </summary>
            /// <param name="wordSize"></param>
            /// <param name="indexValidator">is the hash-key valid, does not check if the entire sequence is valid, only if the small sequence of size wordsize, should be used as hash key</param>
            public RepetitiveSequenceIndexUshort(int wordSize)
            {
                this.wordSize = wordSize;
                this.index = new DNADictionary<ushort>(wordSize);

            }

            public RepetitiveSequenceIndexUshort(DNADictionary<ushort> repetitiveIndex, int wordsize)
            {
                this.index = repetitiveIndex;
                this.wordSize = wordsize;

            }



            /// <summary>
            /// Add a new Sequence to the index
            /// </summary>
            /// <param name="sequence"></param>
            public void AddToIndex(ISequenceContainer sequence)
            {
                for (int i = 0; i < sequence.Length - wordSize; i++)
                {
                    ISequenceContainer seq = sequence.SubSequence(i, wordSize);

                    if (index.ContainsKey(seq))
                    {
                        this.index[seq]++;
                    }
                    else index.Add(seq, 1);

                }

            }


            /// <summary>
            /// Get the DNADictionary which was created for this special instance
            /// </summary>
            public DNADictionary<ushort> Dictionary
            {
                get
                {
                    return index;
                }
            }

            /// <summary>
            /// A repetitive index does only allow for a certain wordsize.
            /// Mixing of the word sizes is not valid. Retrieve the wordsize used in this repetitive index
            /// </summary>
            public int Wordsize
            {
                get
                {
                    return this.wordSize;
                }
            }

            /// <summary>
            /// Get the counts for a certain word
            /// </summary>
            /// <param name="word"></param>
            /// <returns></returns>
            public int? this[ISequenceContainer word]
            {
                get
                {
                    if (index.ContainsKey(word)) return index[word];
                    else return null;
                }
            }







            #region Converters


            /// <summary>
            /// Create the sequenceRepresenting the logOddRepetitiveSequence
            /// </summary>
            /// <param name="sequence"></param>
            /// <param name="generalInfo"></param>
            /// <returns></returns>
            public ISequenceContainer ConvertToLogOdd(ISequenceContainer sequence, INucSeqInfo generalInfo)
            {

                double expectedCount = ((double)generalInfo.Count_ATUCG / Math.Pow(4, wordSize));
                ISequenceContainer logodd = SequenceFactory.GetDefaultSequence();

                ///logOdd=lo
                /// lo=0 -> not overrepresented
                /// lo=1 -> 10 fold overrepresented
                /// lo=2 -> 100 fold overrpresented
                /// etc
                for (int i = 0; i < sequence.Length - wordSize + 1; i++)
                {
                    ISequenceContainer s = sequence.SubSequence(i, wordSize);
                    if (index.ContainsKey(s))
                    {
                        double or = (double)index[s] / expectedCount;
                        if (or < 1) or = 1.0;
                        byte ap = (byte)(Math.Round(Math.Log(or, 10.0), 1) * 10);
#if DEBUG
                        int ind = index[s];
                        double log = Math.Log((double)index[s] / expectedCount, 10.0);
                        double round = (Math.Round(log, 1) * 10);
                        if (ap == (byte)255) throw new Exception("crap");

#endif
                        logodd.Append((byte)ap);
                    }
                    else
                    {
                        logodd.Append((byte)254);
                    }
                }
                for (int i = 0; i < wordSize - 1; i++)
                {
                    logodd.Append((byte)254);
                }

                if (sequence.Length != logodd.Length) throw new DataMisalignedException("the logodd sequence does not have the same size than the original sequence");
                return logodd;

            }


            #endregion

        }


        public class LogRepetitiveSequence: ISimpleSequence,IPositionable, Bio.IO.Seq.IFastaFileReadable<LogRepetitiveSequence>
        {
            /// <summary>
            /// For all repetitive categories: categories are between the lower- and the upper boundary.
            /// Repetitive fractions are assigned to a category when greater or equal than the lower-boundary and small than upper-boundary
            /// lowRep_lowerBoundary smaller_equal lowRep_Region smaller midRep_lowerBoundary
            /// </summary>
            public struct RepetitiveCategoryBoundaries
            {
                public RepetitiveCategoryBoundaries(int lowRep_lowerBoundary, int midRep_lowerBoundary, int highRep_lowerBoundary, int extremeRep_lowerBoundary)
                {
                    this.lowRep_lowerBoundary = lowRep_lowerBoundary;
                    this.midRep_lowerBoundary = midRep_lowerBoundary;
                    this.highRep_lowerBoundary = highRep_lowerBoundary;
                    this.extremeRep_lowerBoundary = extremeRep_lowerBoundary;
                }
                public int lowRep_lowerBoundary;
                public int midRep_lowerBoundary;
                public int highRep_lowerBoundary;
                public int extremeRep_lowerBoundary;
            }

            private ISequenceContainer sequence;
            private string name;
            private string parentName;
            private int? startPosition=null;
            private int? endPosition=null;
            private string comment;
            private bool? isRoot = null;
            private ISequenceValidator validator;


            //
            private RepetitiveCategoryBoundaries boundaries;
            private byte lowRepBoundaryTransformed;
            private byte midRepBoundaryTransformed;
            private byte highRepBoundaryTransformed;
            private byte extremeBoundaryTransformed;

            private int? count_notOR = null;
            private int? count_lowOR = null;
            private int? count_midOR = null;
            private int? count_highOR = null;
            private int? count_extremeOR = null;
            private int? count_invalid = null;
          

            public LogRepetitiveSequence()
            {
                SetDefaultBoundaries();
            }



            public LogRepetitiveSequence(string name, string comment, ISequenceContainer sequence)
            {
                this.name = name;
                this.comment = comment;
                this.sequence = sequence;

                SetDefaultBoundaries();
            }

            public LogRepetitiveSequence(string name, string parentName, ISequenceContainer sequence, int startPosition, int endPosition):this(name,"",sequence)
            {
                this.parentName = parentName;
                this.startPosition = startPosition;
                this.endPosition = endPosition;


            }
            public LogRepetitiveSequence(string name, string parentName, string comment, ISequenceContainer sequence, int startPosition, int endPosition)
                : this(name,parentName,sequence,startPosition,endPosition)
            {

                this.comment = comment;


            }

            /// <summary>
            /// Reset the boundaries to the default values
            /// </summary>
            public void SetDefaultBoundaries()
            {
                RepetitiveCategoryBoundaries b = new RepetitiveCategoryBoundaries();
                b.lowRep_lowerBoundary = 4;
                b.midRep_lowerBoundary = 11;
                b.highRep_lowerBoundary = 51;
                b.extremeRep_lowerBoundary = 201;
                SetNewBoundaries(b);
            }


            private void SetNewBoundaries(RepetitiveCategoryBoundaries boundaries)
            {
                this.boundaries = boundaries;
                this.lowRepBoundaryTransformed = (byte)(Math.Round(Math.Log(boundaries.lowRep_lowerBoundary, 10), 1) * 10.0);
                this.midRepBoundaryTransformed = (byte)(Math.Round(Math.Log(boundaries.midRep_lowerBoundary, 10), 1) * 10.0);
                this.highRepBoundaryTransformed = (byte)(Math.Round(Math.Log(boundaries.highRep_lowerBoundary, 10), 1) * 10.0);
                this.extremeBoundaryTransformed = (byte)(Math.Round(Math.Log(boundaries.extremeRep_lowerBoundary, 10), 1) * 10.0);

                this.count_lowOR = null;
                this.count_notOR = null;
                this.count_midOR = null;
                this.count_highOR = null;
                this.count_extremeOR = null;
                this.count_invalid = null;
            }

            /// <summary>
            /// Gets or sets the name of the present sequence
            /// </summary>
            public string Name
            {
                get
                {

                    return this.name;
                }
                set
                {
                    this.name = value;
                }
            }

            /// <summary>
            /// gets the sequence representing the repetitiveness.
            /// Attention usually it is not obvious how to handle this sequence directly. 
            /// LogRepetitiveSequence encapsulates the basic functionality for handling this sequence
            /// </summary>
            public ISequenceContainer Sequence
            {
                get { return sequence; }
            }
        
            /// <summary>
            /// Gets or sets the comments concerning this sequence
            /// </summary>
            public string Comment
            {
                get
                {
                    return comment;
                }
                set
                {
                    this.comment = value;
                }
            }

            /// <summary>
            /// Gets or sets the validator
            /// </summary>
            public ISequenceValidator Validator
            {
                get
                {
                    return this.validator;
                }
                set
                {
                    this.validator = value;
                }
            }
            /// <summary>
            /// is this sequence valid, using the given validator
            /// </summary>
            /// <returns></returns>
            public bool IsValid()
            {
                return validator.IsValid(sequence);
            }

            /// <summary>
            /// returns the length of the sequence
            /// </summary>
            public long Length
            {
                get
                {
                    return sequence.Length;
                }
            }



            public LogRepetitiveSequence NewInstance(string name, string comment, ISequenceContainer sequence)
            {
                return new LogRepetitiveSequence(name, comment, sequence);
            }

            public LogRepetitiveSequence NewInstance(string name, string parentName, ISequenceContainer sequence, int startPosition, int endPosition)
            {
                return new LogRepetitiveSequence(name, parentName, sequence, startPosition, endPosition);
            }

            public RepetitiveCategory GetRepetitiveCategory(int position)
            {
                byte b = (byte)sequence[position];
                if (b < lowRepBoundaryTransformed) return RepetitiveCategory.None;
                else if (b >= lowRepBoundaryTransformed && b < midRepBoundaryTransformed) return RepetitiveCategory.Low;
                else if (b >= midRepBoundaryTransformed && b < highRepBoundaryTransformed) return RepetitiveCategory.Medium;
                else if (b >= highRepBoundaryTransformed && b < extremeBoundaryTransformed) return RepetitiveCategory.High;
                else if (b >= extremeBoundaryTransformed && b != (byte)244) return RepetitiveCategory.Extreme;
                else if (b == (byte)244) return RepetitiveCategory.Invalid;
                else throw new Exception("Crap");
            }



            public int? Start
            {
                get
                {
                    return this.startPosition;
                }
                set
                {
                    this.startPosition = value;
                }
            }

            public int? End
            {
                get
                {
                    return this.endPosition;
                }
                set
                {
                    this.endPosition = value;
                }
            }

            public bool? IsRoot
            {
                get
                {
                    return isRoot;
                }
                set
                {
                    this.isRoot = value;
                }
            }

            public string ParentName
            {
                get
                {
                    return parentName;
                }
                set
                {
                    this.parentName = value;
                }
            }

            /// <summary>
            /// Gets or sets the boundaries for the repetitive categories
            /// </summary>
            public RepetitiveCategoryBoundaries Boundaries
            {
                get
                {
                    return this.boundaries;
                }
                set
                {
                    SetNewBoundaries(value);
                }
            }

            /// <summary>
            /// Count the regions which are not overrepresented, depending on the specified RepetitiveSubclassBoundaries
            /// </summary>
            public int Count_NotOverrepresented
            {
                get
                {
                    if (this.count_notOR == null)
                    {
                        count_notOR = 0;
                        for (int i = 0; i < sequence.Length; i++)
                        {
                            if (sequence[i] < lowRepBoundaryTransformed) count_notOR++;
                        }
                    }

                    return count_notOR.Value;
                }
            }

            /// <summary>
            /// Count the regions having a low overreprsentation, depending on the specified RepetitiveSubclassBoundaries
            /// </summary>
            public int Count_LowOverrepresented
            {
                get
                {
                    if (this.count_lowOR == null)
                    {
                        count_lowOR = 0;
                        for (int i = 0; i < sequence.Length; i++)
                        {
                            if (sequence[i] >= lowRepBoundaryTransformed && sequence[i] < midRepBoundaryTransformed) count_lowOR++;
                        }
                    }

                    return count_lowOR.Value;
                }
            }

            /// <summary>
            /// Count the regions having a medium overrepresentation, depending on the specified RepetitiveSubclassBoundaries
            /// </summary>
            public int Count_MediumOverrepresentation
            {
                get
                {
                    if (count_midOR == null)
                    {
                        count_midOR = 0;
                        for (int i = 0; i < sequence.Length; i++)
                        {
                            if (sequence[i] >= midRepBoundaryTransformed && sequence[i] < highRepBoundaryTransformed) count_midOR++;
                        }
                    }

                    return count_midOR.Value;
                }
            }

            /// <summary>
            /// Count the regions having a high overrepresentation, depending on the specified RepetitiveSubclassBoundaries
            /// </summary>
            public int Count_HighOverrepresentation
            {
                get
                {
                    if (count_highOR == null)
                    {
                        count_highOR = 0;
                        for (int i = 0; i < sequence.Length; i++)
                        {
                            if (sequence[i] >= highRepBoundaryTransformed && sequence[i] < extremeBoundaryTransformed) count_highOR++;
                        }
                    }

                    return count_highOR.Value;
                }
            }

            /// <summary>
            /// Count the regions having a extreme overrepresentation, depending on the specified RepetitiveSubclassBoundaries
            /// </summary>
            public int Count_ExtremeOverrepresentation
            {
                get
                {
                    if (count_extremeOR == null)
                    {
                        count_extremeOR = 0;
                        for (int i = 0; i < sequence.Length; i++)
                        {
                            if (sequence[i] >= extremeBoundaryTransformed && sequence[i] < (byte)254) count_extremeOR++;
                        }
                    }
                    return count_extremeOR.Value;
                }
            }

            /// <summary>
            /// Count the regions which are invalid
            /// </summary>
            public int Count_Invalid
            {
                get
                {
                    if (count_invalid == null)
                    {
                        count_invalid = 0;
                        for (int i = 0; i < sequence.Length; i++)
                        {
                            if (sequence[i] == (byte)254) count_invalid++;
                        }
                    }

                    return count_invalid.Value;
                }
            }






        }


        public enum RepetitiveCategory
        {
            None,
            Low,
            Medium,
            High,
            Extreme,
            Invalid
        }
    }
    namespace IO
    {

    }
}

/*

        public class RepetitiveIndexBuilder
        {
            private string path;
            private FileStream fs;
            private int wordsize = 0;
            private Bio.Similarity.Hash_ATCG h;
            private ISequenceValidator validat = new DNASequenceValidator_ATCG();


            public RepetitiveIndexBuilder(string path, int wordsize)
            {

                this.path = path;
                fs = new FileStream(path, FileMode.Create, FileAccess.ReadWrite);


                this.wordsize = wordsize;
                this.h = new Bio.Similarity.Hash_ATCG(wordsize);

                InitializeIndex();
            }

            private void InitializeIndex()
            {
                ///The first position is the word size!!!
                fs.Write(new byte[] { (byte)wordsize },0,1);

                //All other positions are the index counts!!!
                for (long i = 0; i < Math.Pow(4.0, wordsize); i++)
                {
                    //fs.Position=i*4; should not be necessary
                    fs.Write(new byte[] { 0, 0, 0, 0 }, 0, 4);
            
                }
                fs.Flush();
            }

            private void IncrementIndex(string index)
            {
                long pos = h.GetHash(index);
                //Multiply the position with 4 and add 1 -> 4 = size of integer; 1 = wordsize
                pos = pos * 4 + 1;
                byte[] b =new byte[4];

                fs.Position = pos;
                fs.Read(b,0,4);
                int i = BitConverter.ToInt32(b, 0);
                i++;
                b = BitConverter.GetBytes(i);
                fs.Position = pos;
                fs.Write(b, 0, 4);
                fs.Flush();

            }

            /// <summary>
            /// Add the specified sequence to the index
            /// </summary>
            /// <param name="addSequence">sequence to add</param>
            public void AddToIndex(ISequence addSequence)
            {

                for (int i = 0; i < addSequence.Length - wordsize; i++)
                {
                    ISequence seq = addSequence.SubSequence(i, wordsize);
                    if (validat.IsValid(seq))
                    {
                        IncrementIndex(seq.ToString());
                    }
                }
            }

            public void Close()
            {
                this.fs.Close();
            }


        }



*/